defmodule Digraph do
  @moduledoc """
  Functions for loading and serializing digraphs to/from Yaml files.
  """

  @type digraph :: tuple()
  @type vertex :: list()
  @type edge :: list()

  @doc """
  Create and return a new digraph containing a few nodes and edges.

  ## Examples

  ```
  iex> g1 = Test03.Digraph.new_digraph()
  ```

  """
  @spec new_digraph() :: digraph()
  def new_digraph() do
    g1 = :digraph.new()
    v1 = add_vertex(g1, :v1)
    v2 = add_vertex(g1, :v2)
    v3 = add_vertex(g1, :v3)
    v4 = add_vertex(g1, :v4)
    :digraph.add_edge(g1, v1, v2, :edge1)
    :digraph.add_edge(g1, v2, v3, :edge2)
    :digraph.add_edge(g1, v1, v4, :edge3)
    g1
  end

  defp add_vertex(digraph, label) do
    vertex = :digraph.add_vertex(digraph)
    vertex1 = :digraph.add_vertex(digraph, vertex, label)
    vertex1
  end

  @doc """
  Show info including vertices and edges about a digraph.

  ## Examples

  ```
  iex> graph5 = Test03.Digraph.new_digraph()
  {:digraph, #Reference<0.190878181.2711224324.220265>,
   #Reference<0.190878181.2711224324.220266>,
   #Reference<0.190878181.2711224324.220267>, true}
  iex> Test03.Digraph.show_digraph(graph5)
  digraph: {:digraph, #Reference<0.190878181.2711224324.220265>,
   #Reference<0.190878181.2711224324.220266>,
   #Reference<0.190878181.2711224324.220267>, true}
  Digraph info: [cyclicity: :cyclic, memory: 1094, protection: :protected]
  Digraph -- No. edges: 3  No. vertices: 4
  Verticies:
      vertex: v4
      vertex: v2
      vertex: v3
      vertex: v1
  Edges:
      edge: v1 --> v2
      edge: v2 --> v3
      edge: v1 --> v4
  :ok
  ```

  """
  @spec show_digraph(digraph()) :: :ok
  def show_digraph(digraph) do
    vertices = :digraph.vertices(digraph)
    edges = :digraph.edges(digraph)
    IO.inspect(digraph, label: "digraph")
    IO.inspect(:digraph.info(digraph), label: "Digraph info")
    #IO.inspect(vertices, label: "vertices")
    no_edges = :digraph.no_edges(digraph)
    no_vertices = :digraph.no_vertices(digraph)
    IO.puts("Digraph -- No. edges: #{no_edges}  No. vertices: #{no_vertices}")
    IO.puts("Verticies:")
    Enum.each(vertices, fn vertex ->
      {_, label} = :digraph.vertex(digraph, vertex)
      IO.puts("    vertex: #{label}")
    end)
    #IO.inspect(edges, label: "edges")
    IO.puts("Edges:")
    Enum.each(edges, fn edge ->
      {_, vertex1, vertex2, label} = :digraph.edge(digraph, edge)
      {_, label1} = :digraph.vertex(digraph, vertex1)
      {_, label2} = :digraph.vertex(digraph, vertex2)
      IO.puts("    edge -- #{label}: #{label1} --> #{label2}")
    end)
    :ok
  end

  @doc """
  Show the (labels of the vertices on) the path from one node to another.

  ## Examples

  ```
  iex(40)> Test03.Digraph.show_digraph_vertices_labels(d1, v0, v2)
  path: node n1 --> node n3:
      vertex: node n1
      vertex: node n2
      vertex: node n3
  :ok
  ```

  """
  @spec show_digraph_vertices_labels(digraph(), vertex(), vertex()) :: :ok
  def show_digraph_vertices_labels(digraph, vertex1, vertex2) do
    {_, v1label} = :digraph.vertex(digraph, vertex1)
    {_, v2label} = :digraph.vertex(digraph, vertex2)
    IO.puts(~s/path: "#{v1label}" --> "#{v2label}":/)
    :digraph.get_short_path(digraph, vertex1, vertex2)
    #|> IO.inspect(label: "path")
    |> Enum.each(fn v ->
      #vinfo = :digraph.vertex(digraph, v)
      #IO.inspect(vinfo, label: "vinfo")
      {_, label} = :digraph.vertex(digraph, v)
      IO.puts(~s/    vertex: "#{label}"/)
    end)
    :ok
  end

  @doc """
  Load (create) a digraph from a Yaml file.

  ## Examples

  ```
  iex> graph1 = Test03.Digraph.load_from_yaml("junk02a.yaml")
  node_names: ["n1", "n2", "n3", "n4"]
  vertices: [n1: [:"$v" | 0], n2: [:"$v" | 1], n3: [:"$v" | 2], n4: [:"$v" | 3]]
  edges: [[:"$e" | 0], [:"$e" | 1], [:"$e" | 2]]
  {:ok,
   {:digraph, #Reference<0.1474948976.2438856707.116099>,
    #Reference<0.1474948976.2438856707.116100>,
    #Reference<0.1474948976.2438856707.116101>, true}, "graph01",
   "Sample graph one", ["n1", "n2", "n3", "n4"],
   [["n1", "n2"], ["n2", "n3"], ["n1", "n4"]],
   [n1: [:"$v" | 0], n2: [:"$v" | 1], n3: [:"$v" | 2], n4: [:"$v" | 3]],
   [[:"$e" | 0], [:"$e" | 1], [:"$e" | 2]]}
  ```

  """
  @spec load_from_yaml(Path.t()) :: {:ok, digraph()}
  def load_from_yaml(in_file_path) do
    {:ok, data} = YamlElixir.read_from_file(in_file_path)
    #{:ok, name} = Map.fetch(data, "name")
    #{:ok, description} = Map.fetch(data, "description")
    {:ok, yaml_nodes} = Map.fetch(data, "nodes")
    {:ok, yaml_edges} = Map.fetch(data, "edges")
    graph = :digraph.new()
    IO.inspect(yaml_nodes, label: "nodes")
    IO.inspect(yaml_edges, label: "edges")
    vertices = Enum.map(yaml_nodes, fn [node_name, label] ->
      node_name_atom = String.to_atom(node_name)
      vertex1 = :digraph.add_vertex(graph)
      vertex2 = :digraph.add_vertex(graph, vertex1, label)
      {node_name_atom, vertex2}
    end)
    edges = Enum.each(yaml_edges, fn [from_node, to_node, label] ->
      from_node_atom = String.to_atom(from_node)
      to_node_atom = String.to_atom(to_node)
      v1 = Keyword.get(vertices, from_node_atom)
      v2 = Keyword.get(vertices, to_node_atom)
      if not (is_nil(v1) or is_nil(v2)) do
        :digraph.add_edge(graph, v1, v2, label)
      end
    end)
    IO.inspect(vertices, label: "vertices")
    IO.inspect(edges, label: "edges")
    db_vertices = :digraph.vertices(graph)
    db_edges = :digraph.edges(graph)
    IO.inspect(db_vertices, label: "db_vertices")
    IO.inspect(db_edges, label: "db_edges")
#    {:ok, graph, name, description, node_names, edge_descriptions,
#      vertices, edges}
    {:ok, graph}
  end

  @doc """
  Serialize a digraph and write it to a Yaml file given the
  digraph, a name, and a description.

  ## Examples

  ```
  iex> Digraph.write_to_yaml("save01.yaml", digraph1, "digraph-1", "digraph number one", true)
  ```

  """
  @spec write_to_yaml(String.t(), digraph(), String.t(), String.t(), boolean()) :: :ok
  def write_to_yaml(out_file_path, digraph, name, description, force \\ false) do
    {:ok, structure} = digraph_to_structure(digraph, name, description)
    #IO.inspect(structure, label: "structure")
    write_yaml(out_file_path, structure, force)
    :ok
  end

  @doc """
  Write an Elixir data structure (a list or map) to a file.

  Args:
  * path -- The path to a file to be created and written to.
  * data -- An Elixir data structure to be converted to Ymal and written.
  * force -- (optional) If true, overwrite existing file; if false,
    and file exists, do not overwrite and exit with :error.

  ## Examples

  ```
  iex> Test03.Digraph.write_yaml("tmp.yaml", data)
  iex> Test03.Digraph.write_yaml("tmp.yaml", data, true)
  ```

  """
  @spec write_yaml(String.t(), term(), bool()) :: :ok | :error
  def write_yaml(path, data, force \\ false) do
    if File.exists?(path) and not force do
      IO.puts("File #{path} exists.  Use force: true to overwrite.")
      :error
    else
      content = Yamlix.dump(data)
      :ok = File.write(path, content)
      :ok
    end
  end

  @doc """
  Read a yaml file using YamlElixir.  Write a yaml file using Yamlix.

  ## Examples

  ```
  iex> Test03.Digraph.yaml_file_to_yaml_file("Data/test01.yaml", "junk.yaml")
  ```

  """
  @spec yaml_file_to_yaml_file(String.t(), String.t()) :: :ok
  def yaml_file_to_yaml_file(in_file_path, out_file_path) do
    {:ok, content} = YamlElixir.read_from_file(in_file_path)
    content1 = Yamlix.dump(content)
    IO.puts(content1)
    :ok = File.write(out_file_path, content1)
    :ok
  end

  def to_tuples(digraph) do
    vertices = :digraph.vertices(digraph)
    edges = :digraph.edges(digraph)
    converted_vertices = Enum.map(vertices, fn vertex ->
      {_, label} = :digraph.vertex(digraph, vertex)
      {:vertex, label}
    end)
    converted_edges = Enum.map(edges, fn edge ->
      {_, vertex1, vertex2, label} = :digraph.edge(digraph, edge)
      {_, label1} = :digraph.vertex(digraph, vertex1)
      {_, label2} = :digraph.vertex(digraph, vertex2)
      {:edge, label, label1, label2}
    end)
    converted_digraph = {converted_vertices, converted_edges}
    converted_digraph
  end

  @doc """
  Convert a digraph to a data structure suitable for serializing to Yaml.

  ## Examples

  ```
  iex> {:ok, data} = Digraph.digraph_to_structure(d1, "digraph001", "a sample digraph")
  {:ok,
   %{
     "description" => "a sample digraph",
     "edges" => [
       ["n3", "n2", "edge1"],
       ["n2", "n1", "edge2"],
       ["n3", "n0", "edge3"]
     ],
     "name" => "digraph001",
     "nodes" => [
       ["n2", "node n2"],
       ["n0", "node n4"],
       ["n3", "node n1"],
       ["n1", "node n3"]
     ]
   }}
  ```

  """
  @spec digraph_to_structure(digraph(), String.t(), String.t()) :: {:ok, map()}
  def digraph_to_structure(digraph, name, description) do
    vertices = :digraph.vertices(digraph)
    edges = :digraph.edges(digraph)
    converted_vertices = Enum.map(vertices, fn vertex ->
      {[_ | nodeitem], label} = :digraph.vertex(digraph, vertex)
      nodename = "n" <> Integer.to_string(nodeitem)
      [nodename, label]
    end)
    converted_edges = Enum.map(edges, fn edge ->
      {_, vertex1, vertex2, label} = :digraph.edge(digraph, edge)
      {[_ | nodeitem1], _label1} = :digraph.vertex(digraph, vertex1)
      {[_ | nodeitem2], _label2} = :digraph.vertex(digraph, vertex2)
      nodename1 = "n" <> Integer.to_string(nodeitem1)
      nodename2 = "n" <> Integer.to_string(nodeitem2)
      [nodename1, nodename2, label]
    end)
    converted_digraph = %{
      "name" => name,
      "description" => description,
      "nodes" => converted_vertices,
      "edges" => converted_edges,
    }
    {:ok, converted_digraph}
  end

end
